#!/usr/bin/env python3

import typing
import asyncio
import threading
import socket
from math import nan
from forge.tasks import wait_cancelable
from forge.acquisition.instrument.standard import IterativeCommunicationsInstrument
from forge.acquisition.instrument.base import BaseContext, BaseBusInterface, CommunicationsError
from forge.acquisition.instrument.parse import parse_number


class Instrument(IterativeCommunicationsInstrument):
    INSTRUMENT_TYPE = "realtimewind"
    MANUFACTURER = "GML"
    DISPLAY_LETTER = "I"
    TAGS = frozenset({"met", "aerosol", "realtimewind"})

    _TIMEOUT = 30.0

    def __init__(self, context: BaseContext):
        super().__init__(context)

        self.data_WS = self.input("WS")
        self.data_WD = self.input("WD")

        self.instrument_report = self.report(
            *self.variable_winds(self.data_WS, self.data_WD, code=""),
        )

        self._incoming_lock = threading.Lock()
        self._incoming_packet: typing.Optional[bytes] = None
        self._incoming_ready: asyncio.Event = None

    @staticmethod
    def _check_packet(data: bytes) -> bool:
        if not data:
            raise CommunicationsError
        if b"BRW" not in data:
            return False
        return True

    async def _process_packet(self, line: bytes) -> None:
        # "2023-07-31 14:51:32","BRW199",1,117.6,1.1,0,1021.051,9.79,9.68,9.55,20.56,13.73

        line = line.strip()
        fields = line.split(b',')
        try:
            (
                _,  # Date and time
                _,  # Instrument ID
                WS, WD,
                *_,
            ) = fields
        except ValueError:
            raise CommunicationsError(f"invalid number of fields in {line}")

        WS = WS.strip()
        if WS == b"+INF" or WS == b"0":
            WS = nan
        else:
            WS = parse_number(WS)

        WD = WD.strip()
        if WD == b"+INF" or WD == b"0":
            WD = nan
        else:
            WD = parse_number(WD)

        self.data_WS(WS)
        self.data_WD(WD)
        self.instrument_report()

    async def initialize_communications(self) -> bool:
        with self._incoming_lock:
            self._incoming_packet = None
            self._incoming_ready.clear()
        await wait_cancelable(self._incoming_ready.wait(), self._TIMEOUT)
        with self._incoming_lock:
            packet = self._incoming_packet
            self._incoming_packet = None
            self._incoming_ready.clear()

        if not self._check_packet(packet):
            raise CommunicationsError(f"invalid packet")
        await self._process_packet(packet)

        return True

    async def step_communications(self) -> bool:
        # Allow unmatched packets, but time out if we don't eventually get one
        async def process_single():
            while True:
                await self._incoming_ready.wait()
                with self._incoming_lock:
                    packet = self._incoming_packet
                    self._incoming_packet = None
                    self._incoming_ready.clear()
                if not self._check_packet(packet):
                    continue
                await self._process_packet(packet)
                return

        await wait_cancelable(process_single(), self._TIMEOUT)
        return True

    async def run(self) -> typing.NoReturn:
        # Send initial information and state
        await self.emit()

        self._incoming_ready = asyncio.Event()

        class IncomingDatagramProtocol:
            def __init__(self, instrument: "Instrument"):
                self.instrument = instrument

            def connection_made(self, transport):
                self.transport = transport

            def connection_lost(self, exc):
                pass

            def datagram_received(self, data: bytes, addr):
                with self.instrument._incoming_lock:
                    self.instrument._incoming_packet = data
                self.instrument._incoming_ready.set()

        try:
            sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
            try:
                sock.bind(('', 2101))
            except:
                sock.close()
                raise
        except IOError:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.bind(('', 2101))

        transport, _ = await asyncio.get_running_loop().create_datagram_endpoint(
            lambda: IncomingDatagramProtocol(self),
            sock=sock,
        )

        try:
            await super().run()
        finally:
            transport.close()


def main():
    from forge.acquisition.instrument.run import launch
    launch(Instrument)


